#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import time
import threading


def _every(delay, stop_event: threading.Event, task, *args, **kwargs):
    next_time = time.time() + delay
    while True:

        segments = delay // 3
        wait_time_per_segment = (next_time - time.time()) // segments
        for i in range(0, segments):
            time.sleep(max(0, wait_time_per_segment))
            if stop_event.is_set():
                return

        try:
            task(*args, **kwargs)
        except Exception as e:
            # in production code you might want to have this instead of course:
            logging.error(f"Problem while executing repetitive task: {e}")

        # skip tasks if we are behind schedule:
        next_time += (time.time() - next_time) // delay + delay


def every(delay, stop_event: threading.Event, task, *args, **kwargs):
    t = threading.Thread(target=lambda: _every(delay, stop_event, task, *args, **kwargs))
    t.start()
    return t



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base64
import hashlib
import json
import os

import requests
import secrets
import threading
import urllib
import webbrowser
from time import sleep
import jwt
from flask import Flask, request, redirect
from werkzeug.serving import make_server
from lmrtfy import _lmrtfy_auth_dir
from typing import Optional
import logging
from datetime import datetime
from lmrtfy import _lmrtfy_config_dir


def get_cliconfig() -> Optional[dict]:

    do_cache = False

    # TODO: What happens if LOCAL and DEV are defined?
    if "LMRTFY_LOCAL" in os.environ:
        url = "http://127.0.0.1:5000/cliconfig"
    elif "LMRTFY_DEV" in os.environ:
        url = "https://dev-api.lmrt.fyi/cliconfig"
    else:
        url = "https://api.lmrt.fyi/cliconfig"
        do_cache = True

    try:
        config_file = _lmrtfy_config_dir.joinpath('cliconfig.json')

        if do_cache:
            if config_file.exists():
                with open(config_file, 'r') as f:
                    d = json.load(f)
                    if float(d['updated']) > (float(datetime.utcnow().timestamp()) - 3600):
                        logging.debug("Loading cached cliconfig.")
                        return d

        logging.debug("Refreshing cliconfig.")
        r = requests.get(url).json()

        r['updated'] = str(datetime.utcnow().timestamp())

        with open(config_file, 'w') as f:
            json.dump(r, f)

        return r
    # TODO: Exception clause too broad
    except:
        logging.error(f"Could not reach the LMRTFY API at {url}. Please try again in a few minutes."
                      "If the error persists please contact hello@lmrt.fyi.")
        exit(-1)


def auth_url_encode(byte_data):
    return base64.urlsafe_b64encode(byte_data).decode('utf-8').replace('=', '')


def generate_challenge(a_verifier):
    return auth_url_encode(hashlib.sha256(a_verifier.encode()).digest())


def save_token_data(token_data):
    try:
        with open(_lmrtfy_auth_dir.joinpath('token'), 'w') as f:
            json.dump(token_data, f)
    except Exception:
        logging.error(f"Could not save token in {_lmrtfy_auth_dir}.")


def load_token_data() -> dict:

    env_token = os.getenv("LMRTFY_ACCESS_TOKEN", None)
    if env_token:
        return {'access_token': env_token, 'id_token': '', 'refresh_token': ''}

    try:
        with open(_lmrtfy_auth_dir.joinpath('token'), 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        logging.error(f'Could not load token in {_lmrtfy_auth_dir}.')
        return {'access_token': '', 'id_token': '', 'refresh_token': ''}


class ServerThread(threading.Thread):

    def __init__(self, app, host: str, port: int):
        threading.Thread.__init__(self)
        self.srv = make_server(host, port, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()


class LoginHandler(object):
    def __init__(self):

        self.cliconfig = get_cliconfig()

        self.received_callback = False
        self.code = None
        self.error_message = None
        self.received_state = None

        self.verifier = auth_url_encode(secrets.token_bytes(32))
        # TODO: Use https for local redirects as well.
        self.redirect_uri = f"http://{self.cliconfig['auth_listener_host']}:{self.cliconfig['auth_listener_ports'][0]}/callback"

    def callback(self):
        # TODO: Make this look nice.
        message = "Success"
        if 'error' in request.args:
            self.error_message = request.args['error'] + ': ' + request.args['error_description']
            message = "Error"
        else:
            self.code = request.args['code']

        self.received_state = request.args['state']
        self.received_callback = True

        return redirect(self.cliconfig['api_login_success_url'], code=302)

    def login(self) -> bool:

        if self.token_is_valid(load_token_data()['access_token']):
            logging.debug('Valid access token found. Login not necessary.')
            return False

        challenge = generate_challenge(self.verifier)
        state = auth_url_encode(secrets.token_bytes(32))

        app = Flask("LMRTFY Login")

        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

        @app.route('/callback')
        def callback():
            return self.callback()

        server = ServerThread(app, self.cliconfig['auth_listener_host'], int(self.cliconfig['auth_listener_ports'][0]))
        server.start()

        base_url = f"{self.cliconfig['auth_authorize_url']}?"
        url_parameters = {
            'audience': self.cliconfig['auth_audience'],
            'scope': "profile openid offline_access email",  # self.cliconfig['auth_scopes'],
            'response_type': 'code',
            'redirect_uri': self.redirect_uri,
            'client_id': self.cliconfig['auth_client_id'],
            'code_challenge': challenge.replace('=', ''),
            'code_challenge_method': 'S256',
            'state': state
        }

        logging.info('Opening browser window for authentication and login.')
        url = base_url + urllib.parse.urlencode(url_parameters)
        webbrowser.open_new(url)

        while not self.received_callback:
            sleep(1)

        server.shutdown()

        if state != self.received_state:
            logging.error("Inconsistency occurred during login. Please log out of all connections.")
            exit(-1)

        if self.error_message:
            logging.error(f'Login not successful: {self.error_message}')
            exit(-1)

        logging.info('Login successful.')
        return True

    def get_token(self):
        logging.debug('Fetching access token.')

        headers = {'Content-Type': 'application/json'}
        body = {'grant_type': 'authorization_code',
                'client_id': self.cliconfig['auth_client_id'],
                'code_verifier': self.verifier,
                'code': self.code,
                'audience': self.cliconfig['auth_audience'],
                'redirect_uri': self.redirect_uri}
        r = requests.post(self.cliconfig['auth_token_url'], headers=headers, data=json.dumps(body))
        data = r.json()

        if self.token_is_valid(data['access_token']):
            headers = {'Content-type': 'application/json', 'Accept': 'text/plain', "Authorization": f"Bearer {data['access_token']}"}
            r = requests.post(self.cliconfig['api_users_url'], headers=headers, data=json.dumps({'id_token': data['id_token']}))
            logging.info(r.json())
            save_token_data(data)

    def token_is_valid(self, token) -> bool:

        logging.debug('Validating auth token.')
        if not token:
            return False

        try:
            r = requests.get(self.cliconfig['auth_jwks_url'])
            jwks = r.json()
            public_keys = dict()

            for jwk in jwks['keys']:
                kid = jwk['kid']
                public_keys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))

            kid = jwt.get_unverified_header(token)['kid']
            jwt.decode(token, key=public_keys[kid], algorithms=["RS256"], audience=self.cliconfig['auth_audience'],
                       verify=True)

            logging.debug('Auth token accepted.')
            return True

        except jwt.exceptions.InvalidSignatureError:
            logging.error('Invalid auth token signature.')
        except jwt.exceptions.ExpiredSignatureError:
            logging.error('Auth token expired.')
        except jwt.exceptions.InvalidAudienceError:
            logging.error('Invalid token.')
        except jwt.exceptions.InvalidIssuerError:
            logging.error('Invalid token.')
        except jwt.exceptions.InvalidIssuedAtError:
            logging.error('Invalid token.')
        except jwt.exceptions.ImmatureSignatureError:
            logging.error('Invalid token.')
        except jwt.exceptions.InvalidKeyError:
            logging.error('Invalid token.')
        except jwt.exceptions.InvalidAlgorithmError:
            logging.error('Invalid token.')
        except jwt.exceptions.MissingRequiredClaimError:
            logging.error('Invalid token.')
        except jwt.exceptions.InvalidTokenError:
            logging.error('Invalid token.')
        except Exception:
            logging.error('Unspecified token validation error accored.')

        return False

    def logout(self):
        logging.debug("Logging out of LMRTFY.")
        delete_token()

        base_url = f"{self.cliconfig['api_logout_url']}?"
        url_parameters = {
            "client_id": self.cliconfig['auth_client_id']
        }

        url = base_url + urllib.parse.urlencode(url_parameters)
        webbrowser.open_new(url)
        logging.info("Logout successful.")


def delete_token():
    try:
        _lmrtfy_auth_dir.joinpath('token').unlink(missing_ok=True)
        logging.debug("Auth token deleted.")
    except:
        pass



# -*- coding: utf-8 -*-
import enum
import json
import logging
import os
import pathlib
import queue
import subprocess
import tempfile
import time
import uuid
from typing import Optional

import certifi
import jwt
import paho.mqtt.client as mqtt
import requests
import yaml

from lmrtfy.helper import NumpyEncoder
from lmrtfy.login import load_token_data, get_cliconfig
from lmrtfy.runner.timer import every

class RunnerStatus(str, enum.Enum):
    IDLE = "IDLE"
    RUNNING = "RUNNING"
    TERMINATED = "TERMINATED"


class JobStatus(str, enum.Enum):
    UNKNOWN = "UNKNOWN"
    ACCEPTED = "ACCEPTED"
    RUNNING = "RUNNING"
    REJECTED = "REJECTED"
    FAILED = "FAILED"
    RESULTS_READY = "RESULTS_READY"
    SUBMITTED = "SUBMITTED"


def on_mqtt_disconnect(client, userdata, flags, rc, properties=None):
    logging.debug("on_disconnect")
    logging.debug(f"  userdata  : {userdata}")
    logging.debug(f"  flags     : {flags}")
    logging.debug(f"  rc        : {rc}")
    logging.debug(f"  properties: {properties}")

    logging.warning("Disconnected from MQTT broker.")


def on_subscribe(client, userdata, flags, rc, properties=None):
    logging.debug("on_subscribe")
    logging.debug(f"  userdata  : {userdata}")
    logging.debug(f"  flags     : {flags}")
    logging.debug(f"  rc        : {[x.getName() for x in rc]}")
    logging.debug(f"  properties: {properties}")




class Runner(object):
    """
    The Runner class is responsible for listening to incoming jobs and the execution of the
    underlying profile with the
    parameters of the incoming jobs.

    Jobs are managed in a queue.

    The Runner sends status messages into a dedicated topic for this specific client.

    :param broker_url: MQTT broker url
    :type broker_url: str
    :param port: Port to use for MQTT
    :type port: int
    :param profile_path: Path to the user-profile that the runner is connected to
    :type profile_path: pathlib.Path
    """

    def __init__(self, runner_id: str, profile_id: str, broker_url: str, port: int, profile_path: pathlib.Path):
        """
        Constructor method
        """
        self.job_list = queue.Queue()
        self.busy = False

        # TODO: test if profile is available
        with open(profile_path, "r") as p:
            self.profile = yaml.safe_load(p)

        logging.debug(self.profile)

        self.filehash = self.profile["filehash"]
        logging.debug(f"Running for profile-id: {self.filehash}")

        # TODO: client id is ideally the same a the filehash I guess. Issue #6
        self.client_id = runner_id
        self.profile_id = profile_id
        # TODO: runner status => Load, RAM Status, ...
        self.runner_status_topic = f"status/runner/{self.client_id}"
        self.heartbeat_topic = f"heartbeat/runner/{self.client_id}"
        self.job_status_topic = f"status/job/"
        # TODO: second status topic for job status

        access_token = load_token_data()['access_token']
        user_id = jwt.decode(access_token, options={"verify_signature": False})["sub"]
        self.user_id = user_id.replace('|','-----')
        # job_topic =  f"$share/{user_id}/{user_id}/{self.filehash}/job"
        # self.client.subscribe(job_topic, qos=2)

        self.client = mqtt.Client(
            client_id=self.client_id,
            protocol=mqtt.MQTTv5,
            transport="websockets"
        )

        access_token = ''
        try:
            access_token = load_token_data()['access_token']
        except Exception:
            # Todo: give error
            pass

        self.client.tls_set(certifi.where())
        # self.client.enable_logger()
        self.client.username_pw_set(self.client_id, access_token)
        self.client.on_connect = self.on_mqtt_connect
        self.client.on_disconnect = on_mqtt_disconnect
        self.client.on_subscribe = on_subscribe

        # keepalive need to be below the time out of the server (mosquitto => 60s default)
        self.client.connect(broker_url, port, keepalive=30)

        # TODO: how to structure shared-groups and topics?

        self.client.loop_start()
        time.sleep(1)

        self.job_status = {
            "filehash": self.filehash,
            "job_id": None,
            "status": JobStatus.UNKNOWN,
            "message": "",
        }
        self.runner_status = {
            "filehash": self.filehash,
            "runner_id": self.client_id,
            "status": RunnerStatus.IDLE,
            "message": "",
        }

        self.publish_runner_status(RunnerStatus.IDLE, "Started idling.")

    def on_mqtt_connect(self, client, userdata, flags, rc, properties=None):
        logging.debug("on_connect")
        logging.debug(f"  userdata  : {userdata}")
        logging.debug(f"  flags     : {flags}")
        logging.debug(f"  rc        : {rc}")
        logging.debug(f"  properties: {properties}")
        if rc == 0:
            logging.info("Successfully connected to MQTT broker.")

        job_topic = f"$share/{self.user_id}/{self.profile_id}/{self.filehash}"
        self.client.subscribe(job_topic)
        logging.debug(f"Listen for jobs on '{job_topic}'.")

    def on_message(self, client, userdate, msg):
        """
        on_message is a callback for the MQTT client which is executed everytime a message is
        received.
        There are two steps:
            1. publish a status message that a job has been received
            2. add the job to the job queue of the runner class for later execution
        """
        logging.debug("on_message")
        logging.debug(f"  userdate: {userdate}")
        logging.debug(f"  msg: {msg}")
        job = json.loads(msg.payload)
        self.job_list.put(job, block=False)
        self.publish_job_status(JobStatus.ACCEPTED, "Accepted job", job["user_id"], json.loads(msg.payload)["job_id"])

    def publish_job_status(self, status: JobStatus, message: str, user_id: str, job_id: Optional[str] = None):
        """
        Publishes a status message with JobStatus and a message to MQTT.

        :param status: One of the available job stati for job monitoring.
        :type status: JobStatus
        :param message: A human-readable message.
        :type message: str
        :param job_id: job_id for the status message. Does not always apply. default: None
        :type job_id: Optional[int]
        """
        # TODO: Why is job_status a class member?
        self.job_status["status"] = status
        self.job_status["message"] = message
        self.job_status["job_id"] = job_id
        self.job_status["user_id"] = user_id

        job_status_topic = f"status/job/{job_id}"
        logging.info(
            f"status update for job '{job_id}': '{status}' with '{message}' for "
            f"profile-id '{self.filehash}' and client-id `{self.client_id}`."
        )
        self.client.publish(job_status_topic, json.dumps(self.job_status))

    def publish_runner_status(self, status: RunnerStatus, message: str):
        if message != "Heartbeat":
            logging.info(f"status update for runner `{self.client_id}: '{status}' with '{message}'.")
        else:
            logging.debug(f"status update for runner `{self.client_id}: '{status}' with '{message}'.")

        self.runner_status["user_id"] = self.user_id
        self.runner_status["status"] = status
        self.runner_status["message"] = message
        self.client.publish(self.runner_status_topic, json.dumps(self.runner_status))


    def execute(self, job_id: str, user_id: str, job_input: dict):
        """
        `execute` is responsible for the actual execution of the command with the correct input
        parameters.

        :param job_id:
        :param job_input:
        """

        config = get_cliconfig()
        command = self.profile["language"]
        if "python" not in command:
            logging.error(f"{command} is currently not supported.")

        script = self.profile["filename"]

        # TODO: Techdebt. Ignoring the cleanup errors is due to a python uptream bug in tempfile. bugs.python.org #43153
        try:
            tempdir = tempfile.TemporaryDirectory(ignore_cleanup_errors=True)  # pylint: disable=consider-using-with
        except:
            tempdir = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        # TODO: Add job_id to path
        os.environ["LMRTFY_TMP_DIR"] = tempdir.name
        logging.debug(f"Set LMRTFY_TMP_DIR to '{tempdir.name}'.")

        # Set write input files
        for var, meta_data in self.profile["variables"].items():
            var_path = pathlib.Path(tempdir.name).joinpath(f"lmrtfy_variable_{var}.json")
            with open(var_path, "w") as input_file:
                json.dump({var: job_input[var]}, input_file, cls=NumpyEncoder)

            logging.debug(f"Created '{str(var_path)}' with expected type '{meta_data['dtype']}'.")
            logging.debug(f"  Set to: '{job_input[var]}'")

        command_args_list = [command, script]
        logging.info(f"Running the follow command: {command_args_list}")

        # Set environment variables for execution and run code
        # TODO: How safe is that?
        self.publish_job_status(JobStatus.RUNNING, "Running script.", user_id=user_id, job_id=job_id)
        os.environ["LMRTFY_DEPLOY_LOCAL"] = "1"
        result_code = subprocess.run(command_args_list, capture_output=True)

        # check results
        if result_code.returncode != 0:
            self.publish_job_status(JobStatus.FAILED, "FAILED!", user_id=user_id, job_id=job_id)
            logging.error("Execution failed for some reason.")
            logging.error(f"stderr: \n {result_code.stderr.decode()}")
        else:
            logging.debug("Execution succeeded.")
            logging.debug(f"stdout of execution: {result_code.stdout.decode()}")

        results = {}
        failed = False
        for res, meta_data in self.profile["results"].items():
            res_path = pathlib.Path(tempdir.name).joinpath(f"lmrtfy_result_{res}.json")
            try:
                with open(res_path, "r") as result_file:
                    results[res] = json.load(result_file)
                    # TODO: add auth code, change results endpoint!
                    files = {'results_file': open(res_path, 'rb')}

                    try:
                        token = load_token_data()['access_token']
                        headers = {"Authorization": f"Bearer {token}"}
                    except Exception:
                        # TODO: Fix exception and log meaningful error message
                        pass

                    r = requests.post(f"{config['api_results_url']}/{user_id}/{job_id}", files=files,
                                      headers=headers)
                    logging.debug(r)
                    if r.status_code != 200:
                        logging.error("Results could not be uploaded")
                        failed = True
            except FileNotFoundError:
                failed = True
                logging.error(f"Result file '{res_path}' not found.")
                break

        if failed:
            self.publish_job_status(JobStatus.FAILED, "Job failed.", user_id, job_id=job_id)
        else:
            self.publish_job_status(JobStatus.RESULTS_READY, "Results ready.", user_id, job_id=job_id)

        logging.debug(results)

        try:
            tempdir.cleanup()
        except:
            pass

        self.busy = False

    def start_listening(self):
        """
        This method waits for arriving jobs and starts the execution.
        """
        self.client.on_message = self.on_message
        #try:
        # queue.get() blocks until an element is inside the queue.
        # This might not work on Windows which has to be tested
        while True:
            # time.sleep(0.1)
            # TODO: this check might be necessary on windows... if not self.job_list.empty()
            #  and not self.busy:
            # if not self.job_list.empty() and not self.busy:
            # self.busy = True
            self.publish_runner_status(RunnerStatus.IDLE, "Waiting for job in MQTT Topic.")

            # blocks until something is in the queue
            job_id, user_id, _, job_param, _, _ = self.job_list.get().values()
            self.publish_runner_status(RunnerStatus.RUNNING, "Starting execution.")
            logging.debug(f"'{job_param}' for job_id '{job_id}'")
            self.execute(job_id, user_id, job_param)
            self.job_list.task_done()
        #except KeyboardInterrupt:
        #    logging.info("Detected KeyboardInterrupt. Stop program.")
        #    self.client.loop_stop()
        #    exit(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import threading
from pathlib import Path
import requests
import yaml
import json
import logging
import time
import atexit
import shortuuid

from lmrtfy import _lmrtfy_profiles_dir
from lmrtfy.runner import Runner, save_json_template, RunnerStatus
from lmrtfy.runner.timer import every
from lmrtfy.login import load_token_data, LoginHandler, get_cliconfig


def send_termination(runner: Runner):
    runner.publish_runner_status(RunnerStatus.TERMINATED, "Terminated Runner")
    time.sleep(0.5)
    runner.client.loop_stop()


def send_heartbeat(runner: Runner):
    runner.publish_runner_status(runner.runner_status["status"], "Heartbeat")


def stop_heartbeat(stop_event: threading.Event, t):
    stop_event.set()
    t.join()


def main(script_path: Path, namespace: str):

    config = get_cliconfig()
    _script_path = str(Path(script_path).resolve())

    _script_identifier = _script_path.replace('/', '_').replace('\\', '_').replace('.', '_')
    _lmrtfy_profile_filename = _lmrtfy_profiles_dir.joinpath(f'{_script_identifier}.yml')

    runner_id = f"local_runner-{shortuuid.ShortUUID().random(length=9)}"

    headers = {'Content-type': 'application/json', 'Accept': 'text/plain',
               "Authorization": f"Bearer {load_token_data()['access_token']}"}
    r = requests.post(config['api_users_url'], headers=headers,
                      data=json.dumps({'id_token': load_token_data()['id_token']}))
    logging.info(r.json())
    nickname = r.json()['nickname']

    if not namespace:
        namespace = nickname
    else:
        namespace = f"{nickname}-{namespace.replace('/','-')}"

    logging.info(namespace)

    url = config['api_catalog_url']
    try:
        with open(_lmrtfy_profile_filename, "r") as p:
            data = {"profile": yaml.safe_load(p), "namespace": namespace, "runner_id": runner_id}

    except FileNotFoundError:
        logging.error(f"No profile for {script_path} found. Please run first without the lmrtfy cli.")
        exit(-1)

    # TODO: This has been caught elsewhere and can be removed/rewritten.
    token = ''
    try:
        token = load_token_data()['access_token']
    except Exception:
        h = LoginHandler()
        if h.login():
            h.get_token()
        try:
            token = load_token_data()['access_token']
        except Exception:
            logging.error('No auth token found. Authentication failed.')
            exit(-1)

    headers = {'Content-type': 'application/json', 'Accept': 'text/plain', "Authorization": f"Bearer {token}"}
    r = requests.post(url, data=json.dumps(data), headers=headers)

    profile_id = None
    if r.status_code == 202 or r.status_code == 200:
        profile_id = r.json()['profile_id']
        logging.info(f"Profile_id to be used for requests: {profile_id}")

        try:
            url = config['api_profiles_url'] + f'/input_template/{profile_id}'
            rr = requests.get(url, headers=headers)
            save_json_template(profile_id, rr.json())
        except:
            logging.error('Could not get input template.')

    else:
        logging.debug(r.content)
        logging.error('Deployment failed.')
        exit(-1)

    r = Runner(runner_id=runner_id,
               profile_id=profile_id,
               broker_url=config['broker_url'],
               port=int(config['broker_port']),
               profile_path=_lmrtfy_profile_filename)

    stop_event = threading.Event()
    t = every(30, stop_event, send_heartbeat, r)

    atexit.register(send_termination, r)

    try:
        r.start_listening()

    except KeyboardInterrupt:
        stop_heartbeat(stop_event, t)
        exit(-1)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import re
import numpy as np


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class Context(object):

    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass


def check_uuid4(test_str: str):
    """
    Checks whether job_od is a valid uuid4 or not.

    returns None if not, otherwise returns re.Match object
    """
    pattern = re.compile("[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")
    return pattern.match(test_str)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import logging
from makefun import create_function
from inspect import Signature, Parameter
import requests

import numpy.typing as npt
from typing import List, Optional, Type
import json
import jwt

import coloredlogs

from lmrtfy.login import LoginHandler, load_token_data, get_cliconfig
from lmrtfy.runner import fetch_template
from lmrtfy.runner import JobStatus
from lmrtfy import _lmrtfy_job_dir
from lmrtfy.helper import NumpyEncoder
from lmrtfy.fetch_results import fetch_results


_log_level = logging.INFO
if 'LMRTFY_DEBUG' in os.environ:
    _log_level = logging.DEBUG

coloredlogs.install(fmt='%(asctime)s [%(process)d] %(levelname)s %(message)s', level=_log_level)


typemap = dict()
typemap['int'] = int
typemap['float'] = float
typemap['complex'] = complex
typemap['string'] = str
typemap['float_array'] = npt.NDArray[float]
typemap['int_array'] = npt.NDArray[int]
typemap['string_array'] = List[str]
typemap['json'] = dict
typemap['bool'] = bool


class Job(object):

    def __init__(self, job_id):
        self.id = job_id
        logging.info(f"Job {self.id} created. Status is {self.status}.")


    @property
    def status(self):
        """
        Queries the job status.

        If it returns `JobStatus.UNKNOWN` the job is likely in failed state.
        """
        try:
            token = load_token_data()['access_token']
            headers = {'Content-type': 'application/json', 'Accept': 'text/plain', "Authorization": f"Bearer {token}"}
            r = requests.get(get_cliconfig()['api_job_url'] + f'/{self.id}', headers=headers)
            status = JobStatus(r.json())
        except Exception as e:
            logging.error(e)
            status = JobStatus.UNKNOWN

        try:
            with open(_lmrtfy_job_dir.joinpath(f"{self.id}.job"), "w") as f:
                json.dump({"id": self.id, "status": status}, f)
        except:
            logging.error(f"Could not write local job-file for job {self.id}.")

        return status

    @property
    def results(self) -> Optional[dict]:
        """
        Returns the results if they are ready and `None` otherwise.
        Results are returned as a dictionary:
        ```json
        {
            "<results_name>": result_value
        }
        ```
        """
        return fetch_results(self.id)
        #try:
        #    config = get_cliconfig()
        #    token = load_token_data()['access_token']
        #    headers = {'Content-type': 'application/json', 'Accept': 'text/plain', "Authorization": f"Bearer {token}"}
        #    r = requests.get(config['api_results_url'] + f"/user/{self.id}", headers=headers)
        #    # TODO: Store results locally and delete job_file.
        #    if r.status_code == 200:
        #        return r.json()
        #    else:
        #        logging.error(f"Could not fetch results from server: {r.status_code}")
        #except ConnectionError as e:
        #    logging.error("Could not access results server.")
        #    logging.error(e.strerror)

    @property
    def ready(self):
        """
        Returns `True` if the job has finished successfully and the results are ready to be fetched.
        """
        if self.status == JobStatus.RESULTS_READY:
            return True
        return False


def unique_name(o, name):
    new_name = name

    if hasattr(o, new_name):
        logging.warning(f"Function {new_name} already exists in catalog!")
        suffix = 1

        while True:
            new_name = name + str(suffix)
            if not hasattr(o, new_name):
                break
            suffix += 1

    return new_name


def signature_from_profile(profile):
    parameters = list()
    for v in profile['variables']:
        parameters.append(
            Parameter(v, kind=Parameter.POSITIONAL_OR_KEYWORD, annotation=typemap[profile['variables'][v]['dtype']]))

    results = list()
    for v in profile['results']:
        results.append(typemap[profile['results'][v]['dtype']])
    results = tuple(results)

    sig = Signature(parameters, return_annotation=Optional[Type[Job]])

    return sig, results


def fetch_profile(profile_id):

    try:
        config = get_cliconfig()
        token = load_token_data()['access_token']
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain', "Authorization": f"Bearer {token}"}
        url = config['api_profiles_url'] + f'/{profile_id}'
        rr = requests.get(url, headers=headers)
        if rr.status_code == 200:
            return rr.json()
        else:
            logging.error('Could not fetch template from server.')
    except:
        logging.error('Fetch template request failed.')


class Namespace(object):

    def __init__(self, name):
        self.__name__ = name


class Function(object):
    def __init__(self, pid):
        self.pid = pid

    def __call__(self, *args, **kwargs):
        pass


class Catalog(object):
    """
    The Catalog object provides an interface to deployed functions that you can run from your code.

    Cloud functions are pulled into the catalog by the constructor, which happens during `from lmrtfy
    import catalog`.

    If you want to retrieve newly deployed function, call `catalog.update()`.

    To run a deployed function from the catalog call `catalog.<deployed_function>(*args, **kwargs)`.

    Each function that has been pulled into the catalog is available via the `help()` command.
    """
    def __init__(self):
        h = LoginHandler()
        if h.login():
            h.get_token()

        self.config = get_cliconfig()
        logging.info(self.config)
        self.token = load_token_data()['access_token']
        self.headers = {'Content-type': 'application/json', 'Accept': 'text/plain',
                        "Authorization": f"Bearer {self.token}"}

        self.profiles = None
        self.update()

    def __add_function(self, namespace, name, sig, res_ann, pid, template):

        template = fetch_template(pid)
        def f(**kwargs) -> Job:
            f.pid = pid
            for p in kwargs:
                template['argument_values'][p] = kwargs[p]

            r = requests.post(self.config['api_submit_url'] + f'/{pid}',
                              data=json.dumps(template, cls=NumpyEncoder),
                              headers=self.headers)

            if r.status_code == 200:
                return Job(r.json()['job_id'])
            if r.status_code == 400:
                logging.error(f'Input Error: {r.json()}')
            elif r.status_code == 404:
                logging.error(f"{r.json()}")

        setattr(namespace, name, create_function(sig, f, func_name=name, qualname=f"{namespace.__name__}-{pid}"))

    def update(self):
        """
        Call `update` to update the catalog with newly deployed functions.
        """
        # TODO: get list of namespaces

        r = requests.get(self.config['api_namespaces_url'], headers=self.headers)
        logging.info(r.json())
        namespaces = r.json()['namespaces']

        for n in namespaces:
            names = n.split('/')
            o = self
            nsn = ""
            for name in names:
                nsn += f"{name}/"
                if not hasattr(o, name):
                    setattr(o, name, Namespace(nsn[:-1]))
                o = getattr(o, name)

            try:
                r = requests.get(f"{self.config['api_namespaces_url']}/{n.replace('/','-')}", headers=self.headers)
                functions = r.json()['functions']

                for func in functions:
                    func_name = unique_name(o, functions[func]['name'])
                    self.__add_function(o, func_name, *signature_from_profile(functions[func]), pid=func, template=functions[func])

            except:  # TODO: Except clause too broad!
                logging.error(f"Could not namespace {n} function catalog.")

    def create_namespace(self, namespace: Namespace, name: str) -> bool:
        """
        Create a unique namespace that collects functions and can be shared.

        Example:
        ```python
        catalog.create_namespace(catalog.<namespace>, "new_namespace")
        ```

        :param namespace: parent namespace
        :param name: Name of the new namespace.
        :return: Returns the `True` on success and `False` in case of an error.
        """
        try:
            new_name = f"{namespace.__name__}/{name}".replace('/', '-')
            r = requests.post(f"{self.config['api_namespaces_url']}",
                              data=json.dumps({"namespace": new_name}), headers=self.headers)
            if r.status_code == 201:
                self.update()
                return True
        except:
            pass

        logging.error(f"Could not create namespace {name}.")

        return False

    def add_function_to_namespace(self, namespace: Namespace, function: Function):
        """
        Add a function to an existing namespace.

        Example:
        ```python
        catalog.add_function_to_namespace(catalog.<namespace>, catalog.<namespace>.<function>
        ```
        """
        # PUT
        r = requests.put(f"{self.config['api_namespaces_url']}/{namespace.__name__.replace('/', '-')}",
                         data=json.dumps({'function': function.__qualname__.split('-')[-1]}), headers=self.headers)
        if r.status_code == 202:
            self.update()
            return True

        return False

    def delete_namespace(self, namespace: Namespace) -> bool:
        """
        Deletes the entire namespace. Deletion will fail if there are still functions in the
        namespace.
        """
        # DELETE
        r = requests.delete(f"{self.config['api_namespaces_url']}/{namespace.__name__.replace('/', '-')}",
                            headers=self.headers)
        if r.status_code == 202:
            self.update()
            return True

        return False

    def remove_function_from_namespace(self, function: Function) -> bool:
        """
        Delete `function` from its namespace.
        """
        # PATCH
        namespace = function.__qualname__.split('-')[0]
        r = requests.patch(f"{self.config['api_namespaces_url']}/{namespace.replace('/', '-')}",
                         data=json.dumps({'function': function.__qualname__.split('-')[-1]}), headers=self.headers)
        if r.status_code == 202:
            self.update()
            return True

        return False

    def share_namespace(self, namespace: Namespace, recipient_email: str) -> Optional[str]:
        """
        Share a namespace with someone else by email. The invite email will be sent to `recipient_email`.
        This does not have to be the email address associated with the account of the person you
        want to share the namespace with.

        Invites only work once.
        """
        id_token = load_token_data()['id_token']
        sender_email = jwt.decode(id_token, options={"verify_signature": False})["email"]
        r = requests.post(f"{self.config['api_invites_url']}",
                          data=json.dumps({"namespace": namespace.__name__,
                                           "recipient_email": recipient_email,
                                           "sender_email": sender_email}), headers=self.headers)

        if r.status_code == 202:
            return r.json()["invite_id"]

    def accept_invite(self, invite_id) -> bool:
        r = requests.get(f"{self.config['api_invites_url']}/{invite_id}", headers=self.headers)
        if r.status_code == 202:
            self.update()
            return True

        return False


catalog = Catalog()



import logging
import json
from typing import Optional
import requests

from lmrtfy import _lmrtfy_template_dir
from lmrtfy.login import load_token_data, get_cliconfig


def fetch_template(profile_id):

    config = get_cliconfig()

    try:
        token = load_token_data()['access_token']
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain', "Authorization": f"Bearer {token}"}
        url = config['api_profiles_url'] + f'/input_template/{profile_id}'
        rr = requests.get(url, headers=headers)
        if rr.status_code == 200:
            return rr.json()
        else:
            logging.error('Could not fetch template from server.')
    except:
        logging.error('Fetch template request failed.')


def save_json_template(profile_id, template):

    try:
        with open(_lmrtfy_template_dir.joinpath(f'{profile_id}.json'), 'w') as f:
            json.dump(template, f)
    except Exception:
        logging.error(f"Could not save template for profile {profile_id} in {_lmrtfy_template_dir}.")


def load_json_template(profile_id) -> Optional[dict]:

    try:
        with open(_lmrtfy_template_dir.joinpath(f'{profile_id}.json'), 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        logging.error(f'Could not load template for profile {profile_id} in {_lmrtfy_template_dir}.')
        logging.info(f'Trying to fetch template for {profile_id}.')
        template = fetch_template(profile_id)
        if template:
            save_json_template(profile_id, template)
            return template

        exit(-1)


from lmrtfy.runner.runner import Runner
from lmrtfy.runner.runner import JobStatus, RunnerStatus
from lmrtfy.runner.__main__ import main


# -*- coding: utf-8 -*-

import sys
import os
import pathlib
import json
import logging
from typing import Union
from lmrtfy.helper import NumpyEncoder
import coloredlogs
from datetime import datetime
from hashlib import sha1

import filehash
import yaml
import numpy as np

from .. import _lmrtfy_profiles_dir

# TODO: talk about file naming convention!
# TODO: Hacky hack. This should be changed.
_script_path = None
if sys.argv:
    if len(sys.argv) > 0:
        if len(sys.argv[0]) > 0:
            _script_path = pathlib.Path(sys.argv[0]).resolve()
            if len(sys.argv[0]) >= len('lmrtfy'):
                if sys.argv[0][-6:] == 'lmrtfy':
                    _script_path = None
                else:
                    coloredlogs.install(fmt='LMRTFY Analyzer %(levelname)s %(message)s')


_run_deployed = False
_tmp_dir = None


if 'LMRTFY_DEPLOY_LOCAL' in os.environ and 'LMRTFY_TMP_DIR' in os.environ:
    logging.info("Running deployed.")
    _run_deployed = bool(int(os.environ['LMRTFY_DEPLOY_LOCAL']))
    _tmp_dir = pathlib.Path(os.environ['LMRTFY_TMP_DIR'])
    logging.info(f"Running: data dir is '{_tmp_dir}'")


if _script_path:
    _sha1hasher = filehash.FileHash('sha1')
    _hash = _sha1hasher.hash_file(_script_path)

    profile = dict()
    profile['language'] = 'python'
    profile['version'] = sys.version
    profile['filename'] = str(_script_path)
    # TODO: Remove all whitespace and things that cannot be a function name from the name
    profile['name'] = profile['filename'].replace('\\', '/').split('/')[-1].split('.py')[0]
    profile['updated'] = datetime.utcnow().timestamp()
    profile['filehash'] = _hash
    profile['variables'] = {}
    profile['results'] = {}


if not _run_deployed:
    if _script_path:
        _script_identifier = str(_script_path).replace('/', '_').replace('\\', '_').replace('.', '_')
        _lmrtfy_profile_filename = _lmrtfy_profiles_dir.joinpath(f'{_script_identifier}.yml')
        _lmrtfy_profile_filename.touch()

if not _run_deployed and _script_path:
    with open(_lmrtfy_profile_filename,'w') as f:
        yaml.dump(profile, f)
        logging.info(f"Wrote profile to {str(_lmrtfy_profile_filename)}.")
        logging.info(f"Using function name: {profile['name']}")


_types = ['json', 'string', 'string_array', 'int', 'int_array', 'float', 'float_array', 'complex',
          'complex_array', 'bool', 'bool_array']
_type_map = {int:'int', bool: 'bool', str: 'string', float: 'float'}
_inverse_type_map = {type(1): int, type(True): bool, type("abc"): str, type(1.1): float,
                     type({}): dict, type(np.zeros(3)): np.asarray, type([]): list}


# a type definition in order to have type hint for variable and result function
supported_object_type = Union[int, float, complex, bool, np.ndarray, list, dict]


def _add_to_api_definition(name: str, kind: str, dtype: str, min = None, max = None,
                           unit: str = None):
    if not _run_deployed and _script_path:
        with open(_lmrtfy_profile_filename, 'r') as p:
            profile = yaml.full_load(p)

        if name in profile[f'{kind}s']:
            logging.error(f"{kind.capitalize()} with name '{name}' already exists! "
                          f"Change the 'name' property in the '{kind}' call.")
            logging.error("Profiling failed. Exiting. Fix the errors above ^^^^^.")
            sys.exit(1)

        logging.info(f"Adding {kind} with name '{name}'")
        profile[f'{kind}s'][name] = {}

        if dtype:
            logging.info(f"Adding datatype '{dtype}' for {kind} with name '{name}'")
            profile[f'{kind}s'][name]['dtype'] = dtype
        if min is not None:
            logging.info(f"Adding minimum value '{min}' for {kind} with name '{name}'")
            profile[f'{kind}s'][name]['min'] = min
        if max is not None:
            logging.info(f"Adding maximum value '{max}' for {kind} with name '{name}'")
            profile[f'{kind}s'][name]['max'] = max
        if unit:
            logging.info(f"Adding unit '{unit}' for {kind} with name '{name}'")
            profile[f'{kind}s'][name]['unit'] = unit

        profile[f'{kind}s_hash'] = sha1(json.dumps(profile[f'{kind}s'], sort_keys=True).encode()).hexdigest()

        with open(_lmrtfy_profile_filename, 'w') as f:
            yaml.dump(profile, f)


def _get_type(a: supported_object_type) -> str:
    """
    Returns the type of `a` as a string.

    Supported types: int, float, complex
    :param a: Object to get type of
    :return: type of the object
    :rytpe: string
    """
    if type(a) in [int, str, bool]:
        return _type_map[type(a)]
    try:
        import numpy as np
        for t in ['int', 'float', 'complex']:
            if t in str(type(a)):
                return t
        if type(a) in [np.ndarray, ]:
            for t in ['int', 'float', 'complex', 'bool']:
                if t in str(a.dtype):
                    return f"{t}_array"
    except ImportError:
        logging.warning("Numpy is not available, so I'm assuming you don't need it "
                        "either.")

    if type(a) in [list, tuple, set]:
        if len(a) > 0:
            if len(set([type(el) for el in a])) > 1:
                return "json"
            for t in ['int', 'float', 'complex', 'bool']:
                if t in str(type(a[0])):
                    return f"{str(t)}_array"
            if 'str' in str(type(a[0])):
                return "string_array"
        return "json"

    if type(a) == dict:
        return "json"


def variable(value: supported_object_type, name: str, min = None, max = None,
             unit: str = None) -> supported_object_type:
    """
    `variable` defines an input parameter of the script. It is used to create the profile for the
    code.

    `v1 = variable(5, name='var1')` create a variable named 'var1' in the profile. If run with the
    python interpreter `v1` takes the value 5 and can be used in the rest of the code just like
    any other variable.

    :param value: Value that is used if script is run with the python interpreter. Can be any expression.
    :param name: Name of the variable used by the API for job submission.
    :param min: [optional] Minimum value that a numeric variable takes. Checked by API.
    :param max: [optional] Maximum value that a numeric variable takes. Checked by API.
    :param unit: [optional] A string declaring the unit. This will likely change to support ´pynt` units.
    :return: `value`
    """
    _add_to_api_definition(name, kind='variable', dtype=_get_type(value), min=min, max=max, unit=unit)

    if _run_deployed and _tmp_dir and _script_path:
        # TODO: warn and except if something fails.
        with open(str(_tmp_dir.joinpath(f'lmrtfy_variable_{name}.json')), 'r') as f:
            tmp = json.load(f)
            # TODO: Make it work for numpy dtypes.
            dtype = _inverse_type_map[type(value)]
            value = dtype(tmp[name])
            logging.info(f"Running: Loaded variable '{name}' of type '{dtype}' and value '{value}'.")

    return value


# TODO: argument ordering is not the same as in variable Issue#7
def result(value: supported_object_type, name: str, min = None, max = None,
           unit: str = None) -> supported_object_type:
    """

    `result` defines an input parameter of the script. It is used to create the profile for the
    code.

    `r1 = result(5*v1, name='res1')` creates a result named 'res1' in the profile. If run with the
    python interpreter `r1` is equal to `5*v1`

    :param value: Value that is used if script is run with the python interpreter. Can be any expression.
    :param name: Name of the variable used by the API for job submission.
    :param min: [optional] Minimum value that a numeric result takes. Currently not checked.
    :param max: [optional] Maximum value that a numeric result takes. Currently not checked.
    :param unit: [optional] A string declaring the unit. This will likely change to support ´pynt` units.
        This is currently not checked and saved but will be used in the future.
    :return: `value`
    """

    _add_to_api_definition(name, kind='result', dtype=_get_type(value), unit=None, min=min, max=max)

    if _run_deployed and _tmp_dir and _script_path:
        # TODO: warn and except if something fails.
        with (open(str(_tmp_dir.joinpath(f'lmrtfy_result_{name}.json')), 'w')) as f:
            json.dump({name: value}, f, cls=NumpyEncoder)
            logging.info(f"Running: Saved result '{name}' with value '{value}'.")

    return value


# TODO: remove the underscore when implemented. The underscore keeps the function from popping up in
#the API reference
def _resource(resource):
    """ This is a filepath, address or URL.

    This is currently not implemented-

    :param resource:
    :return:
    """
    return resource



# -*- coding: utf-8 -*-

import os
import pathlib

__version__ = "0.0.9"
# Don't import or use the underscored variables. These are subject to change.
_user_home = pathlib.Path.home()

# TODO: respect system standard directories, for linux this is:
#  https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
#  issue #2
_lmrtfy_dir = _user_home.joinpath('.lmrtfy')
_lmrtfy_profiles_dir = _lmrtfy_dir.joinpath('profiles')
_lmrtfy_config_dir = _lmrtfy_dir.joinpath('config')
_lmrtfy_auth_dir = _lmrtfy_dir.joinpath('auth')
_lmrtfy_template_dir = _lmrtfy_dir.joinpath('templates')
_lmrtfy_job_dir = _lmrtfy_dir.joinpath('jobs')

if not _lmrtfy_dir.is_dir():
    os.mkdir(_lmrtfy_dir)

if not _lmrtfy_profiles_dir.is_dir():
    os.mkdir(_lmrtfy_profiles_dir)

if not _lmrtfy_config_dir.is_dir():
    os.mkdir(_lmrtfy_config_dir)

if not _lmrtfy_auth_dir.is_dir():
    os.mkdir(_lmrtfy_auth_dir)

if not _lmrtfy_template_dir.is_dir():
    os.mkdir(_lmrtfy_template_dir)

if not _lmrtfy_job_dir.is_dir():
    os.mkdir(_lmrtfy_job_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import fire
import pathlib
import requests
from lmrtfy.runner import load_json_template
import json
import logging
from pathlib import Path
from packaging import version

import coloredlogs

import lmrtfy
import lmrtfy.runner
from lmrtfy.login import LoginHandler, load_token_data, get_cliconfig
from lmrtfy.runner import load_json_template
from lmrtfy.fetch_results import fetch_results


_log_level = logging.INFO
if 'LMRTFY_DEBUG' in os.environ:
    _log_level = logging.DEBUG

coloredlogs.install(fmt='%(asctime)s [%(process)d] %(levelname)s %(message)s', level=_log_level)


def check_version():
    installed_version = version.parse(str(lmrtfy.__version__))
    logging.info(f"Running lmrtfy version {installed_version}.")

    try:
        r = requests.get("https://pypi.org/pypi/lmrtfy/json")
        recent_version = version.parse(str(list(r.json()['releases'].keys())[-1]))

        if installed_version < recent_version:
            logging.warning(f"A new version ({recent_version}) is available. Please run: 'pip install --upgrade lmrtfy'")
    except:
        logging.warning("Version check failed")


class LMRTFY(object):
    """ Let me run that for you.
        Easily deploy your scripts to accept input via a web API.
    """

    def __init__(self):
        logging.info('LMRTFY command line interface')
        check_version()
        self._login_handler = LoginHandler()

    def logout(self):
        """
        Logout from the LMRTFY cloud service
        """
        self._login_handler.logout()

    def login(self):
        """
        Login to the LMRTFY cloud service.
        By using this tool you accept the terms and conditions.
        """

        if "LMRTFY_LOCAL" in os.environ:
            logging.warning("Using the local API.")
        elif "LMRTFY_DEV" in os.environ:
            logging.warning("Using the development API.")

        logging.info('Authenticating for LMRTFY.')
        if self._login_handler.login():
            self._login_handler.get_token()

    def deploy(self, script_path: str, local: bool = False, run_as_daemon: bool = False, namespace: str=""):
        """
        Deploy your script to accept inputs via web-api.

        :param namespace:
        :param script_path: script to be deployed (full path)
        :param local: deployment on this host only (script is executed locally)
        :param run_as_daemon: run local deployment as daemon (in background)
        """

        self.login()

        logging.info(f'Starting deployment of {script_path}')
        if local:

            if run_as_daemon:
                from daemon import DaemonContext as Context
            else:
                from lmrtfy.helper import Context

            logging.warning('Deploying locally.')
            with Context(working_directory='./') as c:
                lmrtfy.runner.main(pathlib.Path(script_path).resolve(), namespace)
        else:
            logging.warning('Deploying to cloud.')
            logging.error("This feature is not yet implemented. Please run 'lmrtfy deploy <script> --local' for now.")

    def submit(self, profile_id: str, input_file: str = None):
        """
        Submit a job to a deployed script via the web.

        :param profile_id: Profile id of the script that you want to run.
        :param input_file: JSON file that serves as the input of the job. Run once without specified
        file to get template.
        """
        self.login()

        if not input_file:
            logging.error('No input file given. Use this template to create one.')
            template = load_json_template(profile_id)
            msg = json.dumps(template, indent=4).split('\n')
            for m in msg:
                logging.warning(m)
            exit(-1)

        try:
            with open(input_file, 'r') as f:
                config = get_cliconfig()
                token = load_token_data()['access_token']
                headers = {'Content-type': 'application/json', 'Accept': 'text/plain', "Authorization": f"Bearer {token}"}
                j = json.load(f)
                d = json.dumps(j)
                logging.info("Submitting job with parameters: ")
                msg = json.dumps(j, indent=4).split('\n')
                for m in msg:
                    logging.warning(m)
                r = requests.post(config['api_submit_url'] + f'/{profile_id}', data=d, headers=headers)
                if r.status_code == 200:
                    logging.info("Job submission successful.")
                    logging.info(f"Job-id: {r.json()['job_id']}")
                else:
                    logging.error("Job submission unsuccessful.")
                    logging.warning(f"Reason: \"{r.text}\"")
        except FileNotFoundError:
            logging.error(f"Opening input file {input_file} failed.")

    def fetch(self, job_id: str, results_dir: str = None):
        """
        Fetch results of a job for a given job id.

        :param results_dir:  Directory where to put the results. Default is './'
        :param job_id: Job id of the job that you want to fetch results for.
        """
        self.login()

        # TODO: These error messages should be created server-side
        #if not check_uuid4(job_id):
        #    logging.error(f"'{job_id}' is not a valid job_id. Please check the output of the "
        #                  f"`lmrtfy submit` step.")
        #    exit(-1)

        logging.info(f'Fetching results for job-id {job_id}')

        results = fetch_results(job_id)

        if results_dir:
            p = Path(results_dir).joinpath(job_id)
        else:
            p = Path('./').joinpath(job_id)

        p.mkdir(exist_ok=True)

        logging.info(f'Writing results in {p.resolve()}')
        for name, content in results.items():
            try:
                logging.info(f"Writing results {name}.")
                with open(p.joinpath(name), "w") as r:
                    json.dump({name: content}, r)
            except IOError:
                logging.error(f"Could not write results {name}")


def main():
    fire.Fire(LMRTFY)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from typing import Optional
import requests
from sys import exit

from lmrtfy.login import load_token_data, get_cliconfig



def fetch_results(job_id: str) -> Optional[dict]:
    try:
        config = get_cliconfig()
        token = load_token_data()['access_token']
        headers = {
            'Content-type': 'application/json',
            'Accept': 'text/plain',
            "Authorization": f"Bearer {token}"
        }
        # TODO: /user/{job_id} is a workaround as of now. not ideal.
        r = requests.get(config['api_results_url'] + f"/user/{job_id}", headers=headers)
        if r.status_code == 200:
            return r.json()
        else:
            logging.error(f"Could not fetch results from server: {r.status_code}")
            logging.error(f"Reason: {r.json()}")

    except ConnectionError as e:
        logging.error("Could not access results server.")
        logging.error(e.strerror)
        exit(-1)
